/*
** Module   : MenuChuiImpl.java
** Abstract : CHUI client implementation of Menu.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 VIG 20150225 Added event processing and drawing methods.
*/
package com.goldencode.p2j.ui.client.chui;

import java.util.List;

import com.goldencode.p2j.ui.Color;
import com.goldencode.p2j.ui.Keyboard;
import com.goldencode.p2j.ui.SubMenuConfig;
import com.goldencode.p2j.ui.WidgetConfig;
import com.goldencode.p2j.ui.WidgetId;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.Dimension;
import com.goldencode.p2j.ui.client.Menu;
import com.goldencode.p2j.ui.client.Point;
import com.goldencode.p2j.ui.client.Rectangle;
import com.goldencode.p2j.ui.client.SubMenu;
import com.goldencode.p2j.ui.client.border.Border;
import com.goldencode.p2j.ui.client.border.LineBorder;
import com.goldencode.p2j.ui.client.chui.driver.ChuiOutputManager;
import com.goldencode.p2j.ui.client.event.EventManager;
import com.goldencode.p2j.ui.client.event.Key;
import com.goldencode.p2j.ui.client.event.KeyInput;
import com.goldencode.p2j.ui.client.event.PaintEvent;
import com.goldencode.p2j.ui.client.widget.AbstractContainer;
import com.goldencode.p2j.ui.client.widget.AbstractWidget;
import com.goldencode.p2j.ui.client.widget.Container;
import com.goldencode.p2j.ui.client.widget.Widget;

/**
 * <code>Menu</code> CHUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class SubMenuChuiImpl
extends SubMenu<ChuiOutputManager>
{
   /** Border around the panel. */
   private Border<ChuiOutputManager> border = null;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenuChuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (SubMenuConfig) cfg);
      
      border = new LineBorder();
      
      //body = new BorderedPanel(new BorderLayout(), border);
   }
   
   /**
    * Drawing worker method.
    */
   public void drawPanel()
   {
      screen().clear(this);

      if (border != null)
         border.drawBorder(this, dimension());
   }

   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      Point p = calcPosition();
      drawTitle(p);

      if (config().displayed)
      {
         drawPanel();
         super.draw();
      }
      
      Dimension menuDim = new Dimension(parent().width(), 1);
      Rectangle menuRect = new Rectangle(parent().location(), menuDim);
      EventManager.postEvent(new PaintEvent(this, menuRect));
   }

   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      // SUB-MENU is not displayed (dropped down) by default
      // setDisplayed(visible);
      
      for(Widget<ChuiOutputManager> w : widgets())
      {
         w.setVisible(true);
      }

      super.setVisible(visible);
   }

   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   @Override
   public Dimension dimension()
   {
      // TODO take note about toggle boxes and submenus. Probably in menu-item or in sub-menu.
      double width = 0; 
      for(Widget<?> w : widgets())
      {
         AbstractWidget<?> aw = (AbstractWidget<?>) w;
         if (width < aw.width())
         {
            width = aw.width(); 
         }
      }
      
      return new Dimension(width + 2, widgets().length + 2); // borders and title
   }

   /**
    * Get widget location relative to container.
    * 
    * @return  relative widget location.
    */
   @Override
   public Point location()
   {
      return calcPosition();
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      // TODO only used for body drawing
      Point screenLoc = super.screenLocation();
      
      return new Point(screenLoc.x + 1, screenLoc.y + 1); // body location
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      boolean menubar = isMenubar();
      if (!config().displayed && !menubar)
      {
         return config().title.length() + 3; // " ->" will be added
      }
      else if (menubar)
      {
         return config().title.length();
      } 
      else // displayed==true  
      {
         return super.width();
      }
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      if (!config().displayed)
         return 1;
      else
         // TODO process this case like MENU ?
         return super.height();
   }
   
   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      int actionCode = event.actionCode();
      if (isMenubar() && 
          super.isEnabled() && 
          (key == Key.VK_ENTER ||
          actionCode == Keyboard.KA_CURSOR_DOWN)) // actionCode == Keyboard.KA_CURSOR_UP
      {
         if (!config().displayed)
         {
            config().displayed = true;
         }
         else
         {
            ThinClient.nextFocus(this.getFocusedChild());
         }
         repaintBody();
         return;
      }
      
      if (actionCode == Keyboard.KA_CURSOR_LEFT)
      {
         ThinClient.previousFocus(this);
         if (isMenubar() && config().displayed)
         {
            hideBody();
         }
         return;
      }

      if (actionCode == Keyboard.KA_CURSOR_RIGHT)
      {
         ThinClient.nextFocus(this);
         if (isMenubar() && config().displayed)
         {
            hideBody();
         }
         return;
      }
      
      // mnemonic
      if (event.keyChar() != KeyInput.CHAR_UNDEFINED)
      {
         parent().processKeyEvent(event);
      }
      else
      {
         super.processKeyEvent(event);
      }
      
   }

   private void repaintBody()
   {
      Rectangle smRect = new Rectangle(screenLocation(), dimension());
      EventManager.postEvent(new PaintEvent(this, smRect));
      repaint();
   }
   
   /**
    * Calculates item position.
    * 
    * @return  new Point of this item.
    */
   private Point calcPosition()
   {
      AbstractContainer<ChuiOutputManager> parent = (AbstractContainer<ChuiOutputManager>) parent();      
      
      List<Widget<ChuiOutputManager>> widgets = parent.getWidgetsList();

      int idx = widgets.indexOf(this);
      
      if (idx == -1) 
      {
         throw new IllegalStateException("MENU-ITEM is not contained in its parents widgets list");
      }
      
      Point p = parent.location();
      
      double x = p.x;
      double y = p.y;
      
      if (parent instanceof Menu && !((Menu<?>)parent).config().popupOnly)
      {
         for(Widget<?> w : widgets)
         {
            if (w.equals(this))
               break;
            x += 1 + w.width();
         }
      }
      else
      {
         x += 1;
         y += 1 + idx;
      }
      
      p = new Point(x, y);
      return p;
   }

   /**
    * Draws menu-item title, drawing mnemonic character as underlined.
    * 
    * @param   p
    *          Title point.  
    * @param   dColor
    *          Color of the title.
    */
   private void drawTitle(Point p)
   {
      Color dColor = hasFocus() ? Color.MESSAGES : Color.NORMAL; // highlighted() ?
      Color mnemonicColor = Color.INPUT;
      
      String title = this.config().title;
      int ampIdx = title.indexOf('&');
      if (ampIdx != -1)
      {
         Point tP = new Point(p);
         String leftPart = title.substring(0, ampIdx);
         screen().at(tP).append(leftPart, dColor);
         
         tP.x += ampIdx;
         String mnemonic = title.substring(ampIdx+1, ampIdx+2);
         screen().at(tP).append(mnemonic, mnemonicColor);
         
         tP.x += 1;
         String rightPart = title.substring(ampIdx+2);
         screen().at(tP).append(rightPart, dColor);
      }
      else
      {
         screen().at(p).append(title, dColor);
      }
      
      if (!isMenubar())
      {
         screen().at(p).append(" ->", dColor);
      }
   }

   /**
    * Check if this sub-menu is right inside MENUBAR (first level)
    * 
    * @return  <code>true</code> if yes.
    */
   private boolean isMenubar()
   {
      Container<ChuiOutputManager> parent = parent();
      return parent instanceof Menu && !((Menu<?>)parent).config().popupOnly;
   }

   private void hideBody()
   {
      config().displayed = false;
      repaintBody();
   }

}
